"""
Contains class representation of executable code block.
"""
from collections import OrderedDict
from prompt_toolkit import prompt
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import style_from_pygments_cls
from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import get_lexer_by_name
from rundoc import BadEnv, CodeFailed, BadInterpreter
import grp
import logging
import os
import pwd
import re
import select
import subprocess
import sys
import time

block_actions = OrderedDict()
def block_action(f):
    """Decorator: Add function as action item in block_actions.

    Function needs to accept exactly 2 arguments:
        args      - List of arguments which is a ':' split of a tag without the
                    first element. First element is used as name of action.
        contents  - Data from the code block.
    """
    block_actions.setdefault(
        f.__name__.replace("_", "-").strip('-'), f)

def fill_env_placeholders(s):
    "Replace %:VARIABLE:% with value of VARIABLE in os.environ."
    variables = re.findall("(%:[A-Za-z_][A-Za-z0-9_]*:%)", s)
    variables = list(set(map(lambda x: x[2:-2], variables)))
    res = s
    for variable in variables:
        res = res.replace("%:"+variable+":%", os.environ.get(variable) or "")
    return res

def __write_file_action(args, contents, mode='a', fill=False):
    "Helper function used by 'create-file' and 'append-file' actions."
    filename    = os.path.expanduser(args.get(0))
    permissions = args.get(1)
    user        = args.get(2)
    group       = args.get(3)
    uid = None
    gid = None
    if user:
        uid = pwd.getpwnam(user).pw_uid
    else:
        uid = os.geteuid()
    if group:
        gid = grp.getgrnam(group).gr_gid
    else:
        if user:
            gid = pwd.getpwuid(uid).pw_gid
        else:
            gid = os.getegid()
    if permissions:
        permissions = int(permissions, 8)
    else:
        permissions = 0o644
    with open(filename, mode) as fh:
        fh.write((fill_env_placeholders(contents) if fill else contents)+"\n")
    os.chmod(filename, permissions)
    os.chown(filename, uid, gid)
    return 0

@block_action
def __create_file(args, contents):
    """create-file:NAME[:OCTAL_PERMISSIONS[:USERNAME[:GROUP]]]
    Create file on path NAME with OCTAL_PERMISSIONS owned by USERNAME:GROUP and
    fill with contents of a code block.
    Example:
        ```create-file:~/.config/test.cfg:640#tag1#tag2
        example = True
        ```
    """
    return __write_file_action(args, contents, 'w+')

@block_action
def __r_create_file(args, contents):
    """r-create-file:NAME[:OCTAL_PERMISSIONS[:USERNAME[:GROUP]]]
    Same as 'create-file' but also replace %:VAR:% placeholders with env vars.
    Example:
        ```env
        example_value=True
        ```
        ```r-create-file:~/.config/test.cfg:640#tag1#tag2
        example = %:example_value:%
        ```
    """
    return __write_file_action(args, contents, 'w+', True)

@block_action
def __append_file(args, contents):
    """append-file:PATH/NAME[:OCTAL_PERMISSIONS[:USERNAME[:GROUP]]]
    Edit file on path NAME, set OCTAL_PERMISSIONS and chown to USERNAME:GROUP
    and paste contents of a code block at the end of the file.
    Example:
        ```append-file:~/.config/test.cfg:640#tag1#tag2
        port = 9200
        ```
    """
    return __write_file_action(args, contents, 'a+')

@block_action
def __r_append_file(args, contents):
    """r-append-file:PATH/NAME[:OCTAL_PERMISSIONS[:USERNAME[:GROUP]]]
    Same as 'append-file' but also replace %:VAR:% placeholders with env vars.
    Note that only code block contents will be replaced. Existing part of file
    will not be changed.
    Example:
        ```env
        port_number=9200
        ```
        ```r-append-file:~/.config/test.cfg:640#tag1#tag2
        port = %:port_number:%
        ```
    """
    return __write_file_action(args, contents, 'a+', True)

@block_action
def __usage(args, contents):
    """usage
    Print code block contents and exit. Used to print usage help if no tags are
    provided.
    Example:
        ```usage
        To setup the product run:
            rundoc run README.md -t setup
        To upgrade the product run:
            rundoc run README.md -t upgrade
        ```
    """
    sys.exit(0)

def get_block_action(tag):
    "Return an action function based on code block tag."
    parts_list = list(filter(bool, tag.split(':')))
    action_name = parts_list.pop(0)
    if action_name not in block_actions:
        return None
    action_args = dict([i, parts_list[i]] for i in range(0, len(parts_list)))
    return lambda contents: block_actions[action_name](action_args, contents)

class DocBlock(object):
    """Single multi-line code block executed as a script.

    Attributes:
        interpreter (str): Interpreter used to run the code.
        code (str): Base code loaded during initialization.
        process (subprocess.Popen): Process object running the interpreter.
        runs (list): List of dictinaries, each containing the following:
            'user_code': User modified version of the code (will be used
                instead of main code unless it's set to None or empty string).
            'output': Full output of executed code block.
            'retcode': exit code of the code block executed
    """
    def __init__(self, code, tags, light=False):
        if light:
            from pygments.styles.manni import ManniStyle as HighlightStyle
            self.HighlightStyle = HighlightStyle
        else:
            from pygments.styles.native import NativeStyle as HighlightStyle
            self.HighlightStyle = HighlightStyle
        interpreter = tags[0]
        self.interpreter = interpreter
        self.code = code
        self.tags = tags
        self.process = None
        self.runs = []  # elements inside are like:
                        #   {
                        #       'user_code':'',
                        #       'output':'',
                        #       'retcode':None,
                        #       'time_start': None,
                        #       'time_stop': None
                        #   }
        self.is_action = interpreter.split(':')[0] in block_actions
        interpreter_exists = subprocess.call(
                ['bash','-c','command -v {} 2>&1>/dev/null'.format(interpreter)]
            ) == 0
        if not self.is_action and not interpreter_exists:
            raise BadInterpreter("Bad interpreter: '{}'".format(interpreter))

    @property
    def last_run(self):
        if len(self.runs):
            return self.runs[-1]
        else:
            return None

    def get_lexer_class(self):
        try:
            # try because lexer may not exist for current interpreter
            return get_lexer_by_name(self.interpreter).__class__
        except:
            # no lexer
            return None

    def get_lexer(self):
        try:
            # try because lexer may not exist for current interpreter
            return PygmentsLexer(self.get_lexer_class())
        except:
            # no lexer
            return None

    def __str__(self):
        lexer_class = self.get_lexer_class()
        code = ''
        if self.last_run:
            code = self.last_run['user_code'].strip()
        else:
            code = self.code
        if lexer_class:
            return highlight(
                code,
                lexer_class(),
                Terminal256Formatter(style=self.HighlightStyle)
                )
        return code

    def get_dict(self):
        return {
            'interpreter': self.interpreter,
            'code': self.code,
            'tags': self.tags,
            'runs': self.runs,
        }

    def prompt_user(self, prompt_text='» '):
        self.last_run['user_code'] = prompt(
            prompt_text,
            default = self.code,
            lexer = self.get_lexer(),
            style = style_from_pygments_cls(self.HighlightStyle),
            )

    def print_output(self, final=False):
        """Read both stdout and stderr, populate them in the variable and print.

        Args:
            final (bool): Used to collect final bytes after the process exists.
        """
        encoding = sys.stdout.encoding
        if final and self.process: # ask for process because might be an action
            line = self.process.stdout.read().decode(encoding)
            self.last_run['output'] += line
            sys.stdout.write(line)
        else:
            assert self.process
            line = self.process.stdout.readline().decode(encoding)
            self.last_run['output'] += line
            sys.stdout.write(line)

    def is_running(self):
        return self.process and self.process.poll() is None

    def run(self, prompt=True):
        if not self.process:
            self.runs.append(
                {
                    'user_code': '',
                    'output': '',
                    'retcode': None,
                    'time_start': None,
                    'time_stop': None,
                }
            )
            if prompt:
                self.prompt_user()
            else:
                self.last_run['user_code'] = self.code
            code = self.last_run['user_code'].strip()
            self.last_run['time_start'] = time.time()
            if self.is_action:
                try:
                    action = get_block_action(self.interpreter)
                    self.last_run['retcode'] = action(code)
                except Exception as e:
                    self.last_run['output'] = str(e)
                    print(str(e))
                    self.last_run['retcode'] = 1
                self.last_run['time_stop'] = time.time()
                self.process = None
                return
            else:
                self.process = subprocess.Popen(
                    [self.interpreter],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    shell=False,
                    )
                self.process.stdin.write(code.encode())
                self.process.stdin.flush()
                self.process.stdin.close()
        while self.is_running():
            self.print_output()
        self.print_output(final=True)
        self.last_run['time_stop'] = time.time()
        self.last_run['retcode'] = self.process.poll()
        self.process = None

    def kill(self):
        if self.process:
             self.process.kill()





"""
Main module for rundoc command line utility.
"""
from rundoc import parsers, clr
from textwrap import dedent
import click
import logging
import rundoc
import sys

logger = logging.getLogger(__name__)

def add_options(options:list):
    "Aggregate click options from a list and pass as single decorator."
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options

_run_control_options = [
    click.option('-s', '--step', default=1, show_default=True,
        help="Step number to start at (greater than 0)."
    ),
    click.option('-p', '--pause', default=0.0, show_default=True,
        help="Number of seconds to pause before each code block."
    ),
    click.option('-r', '--retry', default=0, show_default=True,
        help="Number of retries for failed code blocks."
    ),
    click.option('-P', '--retry-pause', default=1.0, show_default=True,
        help=dedent(
        """Seconds to pause before retriyng failed code block. Actual pause
        will be max value of '--pause' and '--retry-pause'.""")
    ),
    click.option('-o', '--output', type=click.File('w'),
        help=dedent(
        """Output file. All codes, modifications and output of execution will
        be saved here. This file can be later used as an input argument to
        'replay' command.""")
    ),
    click.option('-y', '--yes', is_flag=True,
        help=dedent("""[Deprecated: this is now default behaviour. See '-a' to
        disable] Confirm all steps without prompting user.""")
    ),
    click.option('-a', '--ask', count=True,
        help=dedent("""Let rundoc ask you for input on variables and/or code.
        No option: ask for missing vars.
        -a: ask for all vars.
        -aa: ask for all vars and modify code on failure.
        -aaa: ask for all vars and code."""),
    ),
    click.option('-b', '--breakpoint', type=int, multiple=True,
        help=dedent("""Step number on which to force prompt for code input. You can
        use this option multiple times to add multiple breakpoints."""),
    ),
]

_run_specific_options = [
    click.option('-i', '--inherit-env', is_flag=True,
        help=dedent(
        """Override default env variable values with values from existing
        environment. Useful when you export variables before running rundoc and
        expect them to override default values that are set in the docs.""")
    ),
]

_output_style_options = [
    click.option('--light', is_flag=True,
        help="Use theme for light terminal background."
    ),
]

_tag_options = [
    click.option('-t', '--tags', type=str, default='',
        help=dedent(
        """Hash (#) separated list of tags, e.g. '-t bash#proj-2#test_3'.
        Filter out all code blocks that don't contain at least one of the
        listed tags. (Works like OR logic).""")
    ),
    click.option('-T', '--must-have-tags', type=str, default='',
        help=dedent(
        """Hash (#) separated list of tags, e.g. '-t bash#proj-2#test_3)'.
        Filter out all code blocks that are missing at least one of the listed
        tags. (Works like AND logic).""")
    ),
    click.option('-N', '--must-not-have-tags', type=str, default='',
        help=dedent(
        """Hash (#) separated list of tags, e.g. '-t bash#proj-2#test_3)'.
        Filter out all code blocks that contain at least one of the listed
        tags. (Works like AND NOT logic).""")
    ),
]


@click.group()
@click.version_option(version=rundoc.__version__,
    message="%(prog)s %(version)s - {}".format(rundoc.__copyright__))
@click.option('-d', '--debug', is_flag=True,
    help="Enable debug mode with output of each action in the log.")
@click.pass_context
def cli(ctx, **kwargs):
    logging.basicConfig(
        format = '%(asctime)s.%(msecs)03d, %(levelname)s: %(message)s',
        datefmt = '%Y-%m-%d %H:%M:%S',
        filename = None,
        level = logging.DEBUG if ctx.params.get('debug') else logging.WARNING,
        )

@cli.command()
@add_options(_run_control_options)
@add_options(_output_style_options)
@add_options(_tag_options)
@add_options(_run_specific_options)
@click.argument('input', type=click.File('r'))
def run(**kwargs):
    "Run code from markdown file."
    if kwargs['yes']: print("{}Deprecated option: -y, --yes. See -a, --ask instead.{}".format(clr.yellow, clr.end))
    try:
        commander = parsers.parse_doc(**kwargs)
    except rundoc.BadEnv as e:
        print("{}{}{}".format(clr.red, e, clr.end))
        sys.exit(1)
    try:
        commander.run(**kwargs)
    except KeyboardInterrupt:
        commander.die_with_grace()
        sys.exit(1)
    except rundoc.BadEnv as e:
        print("{}{}{}".format(clr.red, e, clr.end))
        sys.exit(1)
    except rundoc.CodeFailed as e:
        sys.exit(1)

@cli.command()
@add_options(_run_control_options)
@add_options(_output_style_options)
@click.argument('input', type=click.File('r'))
def replay(**kwargs):
    "Run code from the output of 'run' command."
    if kwargs['yes']: print("{}Deprecated option: -y, --yes. See -a, --ask instead.{}".format(clr.yellow, clr.end))
    try:
        commander = parsers.parse_output(**kwargs)
    except Exception as e:
        logger.error('Failed to parse file: {}'.format(e))
        sys.exit(1)
    try:
        commander.run(**kwargs)
    except KeyboardInterrupt:
        commander.die_with_grace()
        sys.exit(1)
    except rundoc.BadEnv as e:
        print("{}{}{}".format(clr.red, e, clr.end))
        sys.exit(1)
    except rundoc.CodeFailed as e:
        sys.exit(1)


@cli.command(name='list-tags')
@click.argument('input', type=click.File('r'))
def list_tags(**kwargs):
    "List all unique tags in the markdown file."
    try:
        tags = parsers.get_tags(**kwargs)
        max_num_len = len(str(tags[0][0]))
        for key, value in tags:
            print("{}{}{}".format(
                value, ' '*(max_num_len - len(str(value))), key))
    except Exception as e:
        logger.error('Failed to parse file: {}'.format(e))
        sys.exit(1)

@cli.command(name='action-tags')
def special_tags(**kwargs):
    "Show available action tags and their use in markdown."
    block_actions = rundoc.block.block_actions
    action_tags_info = ""
    for key in block_actions.keys():
        action_tags_info += "\n" + block_actions[key].__doc__
    print(action_tags_info)
    sys.exit(0)

@cli.command(name='list-blocks')
@add_options(_tag_options)
@add_options(_output_style_options)
@click.option('--pretty', is_flag=True, help="Human readable terminal output.")
@click.argument('input', type=click.File('r'))
def list_blocks(**kwargs):
    "List all blocks that would be executed with selected tags and parameters."
    try:
        parsers.print_blocks(**kwargs)
    except Exception as e:
        logger.error('Failed to parse file: {}'.format(e))
        sys.exit(1)

@cli.command(name='clean-doc')
@click.argument('input', type=click.File('r'))
def clean_doc(**kwargs):
    "Read markdown file, strip any rundoc specific markup and send to stdout."
    try:
        parsers.print_clean_doc(**kwargs)
    except Exception as e:
        logger.error('Failed to parse file: {}'.format(e))
        sys.exit(1)

if __name__ == '__main__':
    cli()




"""
Classes and tools for manipulating code block execution.
"""
from collections import OrderedDict
from prompt_toolkit import prompt
from rundoc import clr, RundocException, BadEnv, CodeFailed, BadInterpreter
from rundoc.block import DocBlock
from time import sleep
import json
import logging
import os
import sys

class OrderedEnv(OrderedDict):
    """Dictionary of environment variables.

    Preserves order of variables as found in string. Lets you load a string of
    variable definitions (new line separated var=value pairs).
    You can append (variable, value) pairs and extend with objects of the same
    type.
    Can prompt user to modify variables. Let's you load variables into os.
    Functions append, extend and import_string contain  'collect_existing_env'
    argument. If you set it to True, then these functions will not overwrite
    existing and os variables in case you set them to empty values (assumes
    that giving empty string as value means you don't with to change those
    variables).
    """
    def __init__(self, title=None):
        super().__init__()
        self.title = title

    def __str__(self):
        return "\n".join([ var+"="+self[var] for var in self ])

    def append(self, var, val, collect_existing_env=True):
        if collect_existing_env:
            val = val or self.get(var) or os.environ.get(var, '')
        self[var] = val

    def extend(self, env, collect_existing_env=True):
        for var in env:
            self.append(var, env[var], collect_existing_env)

    def import_string(self, env_string, collect_existing_env=True):
        for line in env_string.strip().split('\n'):
            if not line: continue
            if '=' not in line:
                raise BadEnv("Bad environment line: {}".format(line))
            var, val = line.split('=', 1)
            var = var.strip()
            val = val.strip()
            if not var:
                raise BadEnv("Bad environment line: {}".format(line))
            self.append(var, val, collect_existing_env)

    def prompt(self):
        if not len(self):
            return
        print(self.title)
        msg = "{}\tConfirm/supply/modify environment variables."
        msg += "\n\tPress Return to finish.{}"
        msg = msg.format(clr.bold, clr.end)
        print(msg)
        env_string = str(self)
        env_string = prompt( '[env]\n', default = env_string )
        self.clear()
        self.import_string(env_string, collect_existing_env=False)

    def prompt_missing(self):
        missing = self.__class__(self.title)
        for var in self:
            if not self[var]:
                missing.append(var, '')
        missing.prompt()
        self.extend(missing, collect_existing_env=False)

    def load(self):
        """
        Set environment according to defined variables.
        """
        for var in self:
            os.environ[var] = self[var]

    def inherit_existing_env(self):
        """
        Override variable values with values from existing environment. Useful
        when you want "outside" environment to have authority over locally
        defined values.
        """
        for var in self:
            val = os.environ.get(var, '') or self.get(var)
            self[var] = val


class DocCommander(object):
    """
    Manages environment and DocBlock objects and executes them in succession.
    """
    def __init__(self):
        self.env = OrderedEnv(
            "\n{}==== env variables{}".format(clr.bold, clr.end)
            )
        self.secrets = OrderedEnv(
            "\n{}==== secrets{}".format(clr.bold, clr.end)
            )
        self.doc_blocks = []
        self.running = False
        self.step = None
        self.output = None

    @property
    def doc_block(self):
        """Current doc_block."""
        if self.step:
            return self.doc_blocks[self.step-1]
            # step-1 because steps start at 0 but are refered to as if they
            # start with 1
        else:
            return None

    def get_dict(self):
        return {
            'env': self.env,
            'code_blocks': [ x.get_dict() for x in self.doc_blocks ]
        }

    def add(self, code, tags, light=False):
        try:
            self.doc_blocks.append(
                DocBlock(
                    code=code,
                    tags=tags,
                    light=light,
                    )
                )
        except RundocException as re:
            logging.error(str(re))
            if self.running:
                self.doc_block.kill()
            sys.exit(1)

    def die_with_grace(self):
        if self.running:
            self.doc_block.kill()
            print("\n==== {}Quit at step {} with keyboard interrupt.{}\n".format(
                clr.red,
                self.step,
                clr.end,
                )
            )
        if self.output:
            self.output.write(
                json.dumps(self.get_dict(), sort_keys=True, indent=4))

    def write_output(self):
        if self.output:
            self.output.write(
                json.dumps(self.get_dict(), sort_keys=True, indent=4))
            print("Output written to: {}".format(self.output.name))

    def run(self,
        step=1, ask=False, breakpoint=[], inherit_env=False, pause=0, retry=0,
        retry_pause=1, output=None, **kwargs):
        """Run all the doc_blocks one by one starting from `step`.

        Args:
            step (int): Number of step to start with. Steps start at 1.
            ask (bool): Ask user to confirm each step.
            breakpoint (list): List of integers representing steps that are
                breakpoints. Rundoc will always prompt user on these steps no
                matter what value `ask` is set to.
            inherit_env (bool): Override env defaults in docs with exported
                values from outside env (for those that exist).
            pause (float): A delay in seconds before the start of each step.
            retry (int): Number of times a step will retry to execute before
                giving up and failing.
            retry_pause (float): Additional pause before retrying same step.
            output (file): Writable file-like object.
        """
        assert self.running == False
        if output is not None:
            self.output = output
        if inherit_env:
            self.env.inherit_existing_env()
            self.secrets.inherit_existing_env()
        if ask>=1:
            self.env.prompt()
            self.secrets.prompt()
        else:
            self.env.prompt_missing()
            self.secrets.prompt_missing()
        msg = "\n{}Running code blocks from supplied documentation."
        msg += "\nModify and/or confirm displayed code by pressing Return.{}"
        print(msg.format(clr.bold, clr.end))
        self.env.load()
        self.secrets.load()
        self.running = True
        self.step = step
        ask_for_prompt_once = False
        while self.step in range(step, len(self.doc_blocks)+1):
            prompt_this_time = \
                ask>=3 or ask_for_prompt_once or self.step in breakpoint
            tags = '[{}] '.format(self.doc_block.interpreter)
            tags += ' '.join(self.doc_block.tags[1:])
            prompt_text = "\n{}=== Step {} {}{}".format(
                clr.bold, self.step, tags, clr.end)
            print(prompt_text)
            if not prompt_this_time:
                print(self.doc_block)
                sleep(pause)
            self.doc_block.run(prompt = prompt_this_time)
            ask_for_prompt_once = False
            if self.doc_block.last_run['retcode'] == 0:
                print("{}==== Step {} done{}\n".format(
                    clr.green, self.step, clr.end))
                self.step += 1
                continue
            # in case it failed:
            self.running = False
            print("==== {}Failed at step {} with exit code '{}'{}\n".format(
                clr.red, self.step, self.doc_block.last_run['retcode'], clr.end))
            if ask>=2:
                msg = "{}{}Press RETURN to try again at step {}.\n"
                msg += "Ctrl+C to quit.{}"
                print(msg.format(clr.red, clr.bold, self.step, clr.end))
                input()
                ask_for_prompt_once = True
                continue
            if len(self.doc_block.runs) > retry:
                self.write_output()
                raise CodeFailed("Failed at step {} with exit code '{}'".format(
                        self.step, self.doc_block.last_run['retcode']))
            print("{}Retry number {}.".format(
                clr.bold, len(self.doc_block.runs), clr.end), end="")
            sleep(retry_pause)
        self.step = 0
        self.write_output()




"""
A command-line utility that runs code blocks from markdown files.
"""

__version__ = "0.3.8"
__license__ = "BSD"
__year__ = "2017-2018"
__author__ = "Predrag Mandic"
__author_email__ = "predrag@eclecticiq.com"
__copyright__ = "Copyright {} {} <{}>".format(
    __year__, __author__, __author_email__)

class clr:
    ''' 
    ANSI colors for pretty output.
    '''
    red = '\033[91m'
    green = '\033[92m'
    blue = '\033[94m'
    yellow = '\033[93m'
    bold = '\033[1m'
    underline = '\033[4m'
    end = '\033[0m'

class RundocException(Exception):
    """Generic rundoc exception."""
    pass

class BadEnv(RundocException):
    """Attempt to load invalid environment name/value"""
    pass

class CodeFailed(RundocException):
    """Raise this when code block returns non-zero exit status."""
    pass

class BadInterpreter(RundocException):
    """Selected interpreter not found or is not executable by current user."""
    pass




#!/usr/bin/env python3

from setuptools import setup, find_packages
import os
import rundoc

def here(*path):
    return os.path.join(os.path.dirname(__file__), *path)

def get_file_contents(filename):
    with open(here(filename), 'r', encoding='utf8') as fp:
        return fp.read()

setup(
    name = 'rundoc',
    description = rundoc.__doc__.strip(),
    long_description=get_file_contents('README.md'),
    url = 'https://gitlab.com/nul.one/rundoc',
    download_url = 'https://gitlab.com/nul.one/rundoc/-/archive/{0}/rundoc-{0}.tar.gz'.format(rundoc.__version__),
    version = rundoc.__version__,
    author = rundoc.__author__,
    author_email = rundoc.__author_email__,
    license = rundoc.__license__,
    packages = [ 'rundoc' ],
    entry_points={ 
        'console_scripts': [
            'rundoc=rundoc.__main__:cli',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Customer Service',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Documentation',
        'Topic :: Software Development :: Documentation',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities',
    ],
    install_requires = [
        'beautifulsoup4>=4.4.1,<5.0',
        'click>=6.7,<7.0',
        'markdown>=2.6.9,<3.0',
        'markdown-rundoc>=0.2.0,<0.3.0',
        'prompt_toolkit>=2.0,<3.0',
        'pygments>=2.2.0,<3.0',
    ],
    python_requires=">=3.4.6",
)




"""
Tools for parsing markdown docs.
"""
from bs4 import BeautifulSoup
from collections import defaultdict
from rundoc.block import DocBlock
from rundoc.commander import DocCommander
import json
import markdown
import markdown_rundoc.rundoc_code
import operator
import re


def parse_doc(input, tags="", must_have_tags="", must_not_have_tags="",
    light=False, **kwargs):
    """Parse code blocks from markdown file and return DocCommander object.

    Args:
        input (file): Readable file-like object pointing to markdown file.
        tags (str): Hash (#) separated list of tags. Markdown code block that
            contain at least one of them will be used.
        must_have_tags (str): Like 'tags' but require markdown code block to
            contain all of them (order not important).
        must_not_have_tags (str): Like 'tags' but require markdown code block
            to contain non of them.
        light (bool): Will use light backgrond color theme if set to True.
            Defaults to False.

    Returns:
        DocCommander object.
    """
    mkd_data = input.read()
    html_data = markdown.markdown(
        mkd_data,
        extensions = [ 
            'markdown_rundoc.rundoc_code(tags={},must_have_tags={}, must_not_have_tags={})'.format(
                tags, must_have_tags, must_not_have_tags,
                )
            ]
        )
    soup = BeautifulSoup(html_data, 'html.parser')
    commander = DocCommander()

    # find blocks
    def is_runnable_block(tag):
        if tag.name != 'code':
            return False
        if 'rundoc_selected' not in tag.get('class', {}):
            return False
        return not bool({
            "env",
            "environ",
            "environment",
            "secret",
            "secrets",
            }.intersection(tag.get('class', {})))
    code_block_elements = soup.findAll(is_runnable_block)
    for element in code_block_elements:
        tags_list = element.get_attribute_list('class')
        tags_list = list(filter(bool, tags_list))
        if tags_list:
            tags_list.remove('rundoc_selected')
            commander.add(element.getText(), tags_list, light)

    # find environments
    def is_environment(tag):
        if tag.name != 'code':
            return False
        if 'rundoc_selected' not in tag.get('class', {}):
            return False
        return bool({'env','environ','environment'}.intersection(
            tag.get('class', {})))
    env_elements = soup.findAll(is_environment)
    env_string = "\n".join([ x.string or '' for x in env_elements ])
    commander.env.import_string(env_string)

    # find secrets
    def is_secret(tag):
        if tag.name != 'code':
            return False
        if 'rundoc_selected' not in tag.get('class', {}):
            return False
        return bool({'secret','secrets'}.intersection(tag.get('class', {})))
    secrets_elements = soup.findAll(is_secret)
    secrets_string = "\n".join([ x.string for x in secrets_elements ])
    commander.secrets.import_string(secrets_string)
    return commander

def parse_output(input, exact_timing=False, light=False, **kwargs):
    """Load json output, create and return DocCommander object.

    Each code block recorded in the otput will be parsed and only code from
    successful attempts will be turned into code blocks for new session. The
    goal is to use original or user modified inputs as a new script.

    Args:
        output (file): Readable file-like object.
        exact_timing (bool): NOT IMPLEMENTED YET!
            If True, a code block will be created for each run try
            and pause between blocks and tries will be calculated from the
            timestamps recorded in the file. The goal is to recreate all exact
            steps that users may have done. Defaults to False.
        light (bool): Will use light backgrond color theme if set to True.
            Defaults to False.

    Returns:
        DocCommander object.
    """
    output_data = input.read()
    data = json.loads(output_data)
    commander = DocCommander()
    for d in data['code_blocks']:
        doc_block = DocBlock(
            code=d['runs'][-1]['user_code'],
            tags=d['tags'],
            light=light,
            )
        commander.doc_blocks.append(doc_block)
    return commander

def get_tags(input):
    tag_dict = defaultdict(int)
    mkd_data = input.read()
    html_data = markdown.markdown(
        mkd_data,
        extensions = [ 'markdown_rundoc.rundoc_code' ]
        )
    soup = BeautifulSoup(html_data, 'html.parser')
    match = re.compile("^.+$")
    code_block_elements = soup.findAll(name='code', attrs={"class":match,})
    for element in code_block_elements:
        class_name = element.get_attribute_list('class')[0]
        if class_name:
            for tag in class_name.split('#'):
                tag_dict[tag] += 1
    sorted_tag_dict = sorted(tag_dict.items(), key=operator.itemgetter(1),
        reverse=True)
    return sorted_tag_dict

def print_blocks(input, tags="", must_have_tags="", must_not_have_tags="",
    light=False, pretty=False):
    commander = parse_doc(input, tags, must_have_tags, must_not_have_tags,
        light)
    if pretty:
        step = 0
        for block in commander.doc_blocks:
            step += 1
            print("{}. [{}] {}".format(step, block.interpreter, '#'.join(block.tags)))
            print("=================")
            print(block)
            print("")
    else:
        print(json.dumps(commander.get_dict(), sort_keys=True, indent=4))

def print_clean_doc(input):
    mkd_data = input.read()
    print(
        re.sub(
            '(\\n```[^#:]*).*?(\\n.*?```\\n)',
            '\\1\\2',
            mkd_data,
            flags=re.DOTALL
            )
        )

